/**
 * hpgv_composite.c
 *
 * Copyright (c) 2008 Hongfeng Yu
 *
 * Contact:
 * Hongfeng Yu
 * hfstudio@gmail.com
 * 
 *
 * All rights reserved.  May not be used, modified, or copied 
 * without permission.
 *
 */

#include "hpgv_composite.h"
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define PARTIAL_IMAGE_MSG   100
#define FULL_IMAGE_MSG      101

/**
 * composite_t
 *
 */
typedef enum composite_t {
    HPGV_TTSWAP
} composite_t;



/**
 * count_t
 *
 */
typedef struct count_t {
    int   key;
    int   count;
} count_t;


/**
 * depth_t
 *
 */
typedef struct depth_t {
    int   mpiid;
    float d;
} depth_t;


/**
 * swap_message_t
 * 
 */
typedef struct swap_message_t {
    uint32_t mpiid;                 /* Remote processor id */
    uint16_t pixelsize;             /* byte size of a pixel */   
    uint64_t recordcount;           /* number of send pixels */
    uint64_t offset;                /* offset of send pixels */ 
    uint8_t  *partial_image;        /* null if it's a send message. */
                                    /* the actual send image is stored in */
                                    /* the swap controller */
    struct swap_message_t *next;    
} swap_message_t;


/**
 * swap_schedule_t
 * 
 */
typedef struct swap_schedule_t {
    uint32_t mpiid;    
    uint32_t curstage;
     
    uint32_t send_count;
    swap_message_t *first_send;
    MPI_Request *isendreqs;
    MPI_Status  *isendstats;    
    uint64_t send_recordcount;
    
    uint32_t recv_count;
    swap_message_t *first_recv;
    MPI_Request *irecvreqs;
    MPI_Status  *irecvstats;    
    uint64_t recv_recordcount;        
    
} swap_schedule_t;


/**
 * array_t
 *
 */
typedef struct array_t{
    int count;
    int cursor;
    uint32_t *array;
} array_t;
        

/**
 * swapnode_t
 *
 */
typedef struct swapnode_t {
    uint32_t workernum;             /* number of processors which is equal to */
                                    /* number of image partition */    
    uint16_t level;                 /* stage in swap process */
    uint32_t childnum;              /* number of children */
    uint32_t firstworker;           /* first processor id */
    uint32_t lastworker;            /* last processor id */
    uint32_t *assignorder;          /* children's assignment order */
    struct swapnode_t *parent;      /* parent */
    struct swapnode_t **children;   /* children */
} swapnode_t;



/**
 * swaptree_t
 *
 */
typedef struct swaptree_t {
    swapnode_t *root;
    uint16_t maxlevel;
} swaptree_t;


/**
 * split_t
 *
 */ 
typedef struct split_t {
    int count;
    int *split;
} split_t;


/**
 * update_t
 * 
 */
typedef struct update_t {
    uint8_t worker;
    uint8_t workersplit;
    uint8_t tree;
    uint8_t order;
    uint8_t schedule; 
} update_t;
    

/**
 * swap_control_t
 * 
 */
typedef struct swap_control_t {
    /* rendering parameters */
    uint32_t image_size_x, image_size_y;
    composite_t composite_type;
    depth_t depth;    
    char prefix[64];
    
    /* mpi */    
    int mpiid, workid;
    int *workid_2_mpiid;    
    int *workers, workernum;
    int totalprocnum;
    MPI_Comm mpicomm;
    int root;
    int roothasimage;    

    /* swap tree */
    swaptree_t *swaptree;
    
    /* schedule responsible for sending and receiving data */    
    swap_schedule_t **schedule;
    uint32_t send_count;
    uint64_t send_recordcount;    
    uint32_t recv_count;
    uint64_t recv_recordcount;
    
    /* the number of swap */
    int stagenum;
    
    /* partial image that I am responsible for. Ping-pong buffer for 
       avoiding to copy my image during multiply stages*/
    uint8_t *partial_image[2];
    
    /* final image, only significant at root */
    uint8_t *final_image;
    
    /* image format */
    int imageformat;
    
    /* image type */
    int imagetype;
    
    /* total pixels of the final image */
    uint64_t totalpixels;

    /* useful image */    
    uint64_t usefulpixels;
    
    /* byte size of a pixel */
    uint16_t pixelsize;
    
    /* total byte size */
    uint64_t totalpixelbytes;
    
    /* assignment of the last stage which is used to 
       gather the final image*/
    int64_t *finalassign;
    
    /* cpu split table */
    split_t **workersplit;
    
} swap_control_t;

static swap_control_t *theSwapControl = NULL;


/**
 * swaptree_delete_treenode
 *
 */
static void 
swaptree_delete_treenode(int mpiid, swapnode_t *swapnode) 
{
    uint32_t i;

    if (swapnode == NULL)
        return;
    
    /* delete leaf */
    if (swapnode->childnum == 0) {
        if (swapnode->assignorder != NULL) {
            free(swapnode->assignorder);
        }
        free(swapnode);
        return;
    }
    
    /* delete children */
    for (i = 0; i < swapnode->childnum; i++) {
        swaptree_delete_treenode(mpiid, swapnode->children[i]);        
    }
    
    /* delete itself */
    if (swapnode->assignorder != NULL ) {
        free(swapnode->assignorder);
    }
    
    if (swapnode->children != NULL) {
        free(swapnode->children);
    }
    
    free(swapnode);
}


/**
 * swaptree_delete_tree
 *
 */
static void 
swaptree_delete_tree(int mpiid, swaptree_t *swaptree)
{
    swaptree_delete_treenode(mpiid, swaptree->root);
        
    free(swaptree);
}


/**
 * swaptree_get_split
 *
 */
static split_t *
swaptree_get_split(int mpiid, int workernum, split_t **wokersplit)
{
    split_t *split = wokersplit[workernum];
    
    HPGV_ASSERT(mpiid, split, "Can not find the split.", HPGV_ERROR);
    
    return split;
}

 
/**
 * swaptree_new_treenode
 *
 */
static void 
swaptree_new_treenode(int mpiid, split_t **wokersplit, 
                      swapnode_t *swapnode, uint16_t *maxlevel)
{   
    uint32_t workernum, i;          
    swapnode_t *child = NULL;
    split_t *childsplit;
 
    workernum = swapnode->workernum;
        
    
    /* stop if it is leaf */
    if (workernum == 1) {
        swapnode->assignorder = (uint32_t *)calloc(swapnode->workernum, 
                                                    sizeof(uint32_t *));
        HPGV_ASSERT(mpiid, swapnode->assignorder, 
                    "Out of memory.", HPGV_ERR_MEM);                        
        swapnode->assignorder[0] = swapnode->firstworker;
        return;        
    }
    
    childsplit = swaptree_get_split(mpiid, workernum, wokersplit);
    
    /* new children */
    swapnode->children = (swapnode_t **)calloc(childsplit->count, 
                          sizeof(swapnode_t *));
    HPGV_ASSERT(mpiid, swapnode->children, "Out of memory.", HPGV_ERR_MEM);    
    
    swapnode->childnum = childsplit->count;
        
    for (i = 0; i < swapnode->childnum; i++) {        
        /* allocate */
        child = (swapnode_t *)calloc(1, sizeof(swapnode_t));
        HPGV_ASSERT(mpiid, child, "Out of memory.", HPGV_ERR_MEM);
        
        /* entry */
        child->workernum = childsplit->split[i];
        child->level = swapnode->level + 1;
        child->parent = swapnode;
        child->childnum = 0;
                
        if (i == 0) {
            child->firstworker = swapnode->firstworker;
        } else {
            child->firstworker = swapnode->children[i-1]->lastworker + 1;
        }
                
        child->lastworker = child->firstworker + child->workernum - 1;
        child->assignorder = NULL;
        
        /* update maximal stage */
        if (*maxlevel < child->level) {
            *maxlevel = child->level;
        }
        
        /* further partition the child if it's not leaf*/         
        if (child->workernum > 1) {
            swaptree_new_treenode(mpiid, wokersplit, 
                                  child, maxlevel);            
        } else {
                        
            child->assignorder = (uint32_t *)calloc(child->workernum, 
                                                sizeof(uint32_t *));
            HPGV_ASSERT(mpiid, child->assignorder, 
                        "Out of memory.", HPGV_ERR_MEM);
                        
            child->assignorder[0] = child->firstworker;
        }
        
        /* attach the child */
        swapnode->children[i] = child;
    }
}


/**
 * swapaux_array_compare
 *
 */
static int 
swapaux_array_compare(const void * a, const void * b)
{
    int aa = ((array_t*)(a))->count;
    int bb = ((array_t*)(b))->count;
    
    if (aa == bb) {
        return 0;
    }
    
    if (aa > bb) {
        return -1;
    }
    
    return 1;
}


/**
 * swaptree_order_assignment: bottom-up recursively order assignment
 *
 */
static void 
swaptree_order_assignment(uint32_t mpiid, swapnode_t *swapnode, 
                          uint16_t maxlevel)
{
    int i, count, alldone;    
    array_t *sorted = NULL;
    
    /* skip if swapnode is NULL */
    if (swapnode == NULL) {
        return;
    }
    
    
    if (swapnode->childnum == 0) {
        return;
        /*
        HPGV_ABORT(mpiid, 
                   "This is a no-leaf node. Child number can not be zero.",
                   HPGV_ERROR);
        */
    }
        
    /* generate assignment of children */
    for (i = 0; i < swapnode->childnum; i++) {
        if (swapnode->children[i]->assignorder == NULL) {
            /* order assignment for the non-leaf node */
            swaptree_order_assignment(mpiid, swapnode->children[i], maxlevel);
                                
        } else {            
            
            /* ===================================================== */
            /* BUG : for qkswap, a tree does not need to be comoplete */
            
            /* this is leaf node. verify if it is a complete tree */
            /*
            if (swapnode->children[i]->level != maxlevel) {
                HPGV_ABORT(mpiid, "It is not a complete tree", HPGV_ERROR);
            }
            */
            
            /* ===================================================== */
        }
    }
    
    /* allcoate the memory for assignment */
    swapnode->assignorder = (uint32_t *)calloc(swapnode->workernum, 
                             sizeof(uint32_t));
    HPGV_ASSERT(mpiid, swapnode->assignorder, "Out of memory", HPGV_ERR_MEM);
    
    /* sort the children according to their number of processors */
    sorted = (array_t *)calloc(swapnode->childnum, sizeof(array_t));
    HPGV_ASSERT(mpiid, sorted, "Out of memory", HPGV_ERR_MEM);
    
    for (i = 0; i < swapnode->childnum; i++) {
        sorted[i].count = swapnode->children[i]->workernum;
        sorted[i].cursor = 0;
        sorted[i].array = swapnode->children[i]->assignorder;
    }
    
    qsort(sorted, swapnode->childnum, sizeof(array_t), swapaux_array_compare);
    
    /* fill children's processors into my assignment list */
    alldone = HPGV_TRUE;
    count = 0;            
    do{
        alldone = HPGV_TRUE;        
        for (i = 0; i < swapnode->childnum; i++) {
            if (sorted[i].cursor < sorted[i].count) {
                swapnode->assignorder[count++] 
                        = sorted[i].array[sorted[i].cursor];    
                sorted[i].cursor++;
                alldone = HPGV_FALSE;
            }                
        }    
    }while(alldone == HPGV_FALSE);    
    
    free(sorted);
}


/**
 * swaptree_new_tree
 *
 */
static swaptree_t * 
swaptree_new_tree(int mpiid, int workernum, split_t **wokersplit)
{
    swapnode_t *root;
    swaptree_t *swaptree;
    uint16_t maxlevel;
    
    /* root */
    root = (swapnode_t *)calloc(1, sizeof(swapnode_t));    
    HPGV_ASSERT(mpiid, root != NULL, "Out of memory.", HPGV_ERR_MEM);

    /* entry */
    root->workernum = workernum;
    root->level = 0;
    root->parent = NULL;
    root->childnum = 0;
    root->firstworker = 0;
    root->lastworker = workernum - 1;
    root->assignorder = NULL;
        
    maxlevel = 0;
    
    /* recursively construct */
    swaptree_new_treenode(mpiid, wokersplit, root, &maxlevel);
    
    /* assignment */
    swaptree_order_assignment(mpiid, root, maxlevel);
    
    /* swap tree */    
    swaptree = (swaptree_t *)calloc(1, sizeof(swaptree_t));
    HPGV_ASSERT(mpiid, swaptree, "Out of memory", HPGV_ERR_MEM);
    
    /* attach root to swap tree*/
    swaptree->root = root;
    swaptree->maxlevel = maxlevel;

    return swaptree;
}


/**
 * swaptree_set_schedule: generate schedule each processor each level
 *
 */
static void 
swaptree_set_schedule(swap_control_t *swapctrl, 
                      swapnode_t *swapnode,                     
                      swap_schedule_t **swapschedule)                            
{
    int mpiid = swapctrl->mpiid;
    int workid = swapctrl->workid;
    int *workid_2_mpiid = swapctrl->workid_2_mpiid;
    uint64_t usefulpixels = swapctrl->usefulpixels;
    uint16_t pixelsize = swapctrl->pixelsize;
    uint32_t maxstage = swapctrl->stagenum;    
    
    int32_t myworkid, mystage;
    uint32_t my_curstart, my_curend, my_prevstart, my_prevend;    
    uint32_t otherworkid, otherstart, otherend;
    uint32_t overlapstart, overlapend;
    uint32_t i, j;
    swap_message_t *newmsg;
    swap_schedule_t *schedule;
    swapnode_t *child = NULL;
        
    my_curstart = my_curend = my_prevstart = my_prevend = 0;
    
    /* skip leaf node */
    if (swapnode->childnum == 0) {
        return;
    }
    
    /* find my current assignment */
    myworkid = -1;
    for (i = 0; i < swapnode->workernum; i++) {        
        if (workid == swapnode->assignorder[i]) {
            myworkid = workid;
            mystage = maxstage - swapnode->level - 1;        
            my_curstart = BLOCK_LOW(i, swapnode->workernum, usefulpixels);
            my_curend = BLOCK_HIGH(i, swapnode->workernum, usefulpixels);
            break;
        }
    }
    
    if (myworkid == -1) {
        
        /* if i am root and do not participate compositing */
        if (mpiid == swapctrl->root && !swapctrl->roothasimage) {
            return;
        }
        
        /* else i should be in the tree */    
        HPGV_ABORT(mpiid, "Can not find processor.", HPGV_ERROR);
    }
        
    /* generate a schedule for the current stage */
    schedule = (swap_schedule_t *)calloc(1, sizeof(swap_schedule_t));
    HPGV_ASSERT(mpiid, schedule, "Out of memory.", HPGV_ERR_MEM);    
            
    swapschedule[mystage] = schedule;
    schedule->mpiid = mpiid;
    schedule->curstage = mystage;

            
    /* scan the children to prepare recv buffer */
    for (i = 0; i < swapnode->childnum; i++) {

        for (j = 0; j < swapnode->children[i]->workernum; j++) {
        
            otherworkid = swapnode->children[i]->assignorder[j];           
            otherstart = BLOCK_LOW(j, swapnode->children[i]->workernum, 
                                   usefulpixels);
            otherend = BLOCK_HIGH(j, swapnode->children[i]->workernum, 
                                  usefulpixels);

            if (otherworkid == myworkid) {
                /* get my previous assignment */
                my_prevstart = otherstart;
                my_prevend = otherend;
                
                HPGV_ASSERT(mpiid, child == NULL, 
                            "I can only be in one child node.", 
                            HPGV_ERR_MEM);
                                
                child = swapnode->children[i];
            }
            
            
            /* there is overlap */
            if ( !(my_curstart > otherend || my_curend < otherstart)) {

                overlapstart = my_curstart >= otherstart ?
                    my_curstart : otherstart;
                overlapend = my_curend <= otherend ?
                    my_curend : otherend;
                
                newmsg = (swap_message_t *)calloc(1, sizeof(swap_message_t));
                HPGV_ASSERT(mpiid, newmsg, "Out of memory,", HPGV_ERR_MEM);

                newmsg->mpiid = workid_2_mpiid[otherworkid];
                newmsg->offset = overlapstart;
                newmsg->recordcount = overlapend - overlapstart + 1;   
                newmsg->pixelsize = pixelsize;                                   
                
                /* prepare buffer to recv the image */
                if (newmsg->mpiid != mpiid) {
                    
                    newmsg->partial_image = 
                            (uint8_t *)calloc(newmsg->recordcount, pixelsize);
                    
                    HPGV_ASSERT(mpiid, newmsg->partial_image,
                                "Out of memory.", HPGV_ERR_MEM);
                                        
                    schedule->recv_count++;
                    schedule->recv_recordcount += newmsg->recordcount;
                }
                    
                newmsg->next = schedule->first_recv;
                schedule->first_recv = newmsg;
            }
        }
    }
    
    /* scan the sibling to prepare send buffer */
    for (i = 0; i < swapnode->workernum; i++) {
        otherworkid = swapnode->assignorder[i];
        
        if (myworkid != otherworkid) {
            otherstart = BLOCK_LOW(i, swapnode->workernum, usefulpixels);
            otherend   = BLOCK_HIGH(i, swapnode->workernum, usefulpixels);
                
            /* there is overlap */
            if ( !(my_prevstart > otherend || my_prevend < otherstart)) {
                overlapstart = my_prevstart >= otherstart ?
                               my_prevstart : otherstart;
                overlapend = my_prevend <= otherend ?
                             my_prevend : otherend;
                
                newmsg = (swap_message_t *)calloc(1, sizeof(swap_message_t));
                HPGV_ASSERT(mpiid, newmsg, "Out of memory.", HPGV_ERR_MEM);

                newmsg->mpiid = workid_2_mpiid[otherworkid];
                newmsg->offset = overlapstart;
                newmsg->recordcount = overlapend - overlapstart + 1;
                newmsg->pixelsize = pixelsize; 
                
                newmsg->next = schedule->first_send;
                schedule->first_send = newmsg;
                schedule->send_count++;
                schedule->send_recordcount += newmsg->recordcount;    
            }
        }
    }
    
    /* allocate structures for asynchronous MPI send/receive */
    schedule->isendreqs = 
        (MPI_Request *)calloc(schedule->send_count, sizeof(MPI_Request));
    HPGV_ASSERT(mpiid, schedule->isendreqs, "Out of memory.", HPGV_ERR_MEM);
    
    schedule->isendstats = 
        (MPI_Status *)calloc(schedule->send_count, sizeof(MPI_Status));
    HPGV_ASSERT(mpiid, schedule->isendstats, "Out of memory.", HPGV_ERR_MEM);

    
    schedule->irecvreqs = 
        (MPI_Request *)calloc(schedule->recv_count, sizeof(MPI_Request));
    HPGV_ASSERT(mpiid, schedule->irecvreqs, "Out of memory.", HPGV_ERR_MEM);
    
    schedule->irecvstats = 
        (MPI_Status *)calloc(schedule->recv_count, sizeof(MPI_Status));
    HPGV_ASSERT(mpiid, schedule->irecvstats, "Out of memory.", HPGV_ERR_MEM);
    
    HPGV_ASSERT(mpiid, child != NULL, "Can not find a child containing myself.",
                HPGV_ERROR);
    
    /* goto lower level */    
    swaptree_set_schedule(swapctrl, child, swapschedule);  
}



/**
 * swapaux_splitworker_ttswap
 *
 */
static void
swapaux_splitworker_ttswap(int mpiid, int workernum, split_t **split)
{
    int count, partition[3], lognum[3], i;
    split_t *newsplit = NULL;
    
    if (workernum == 0) {
        split[workernum] = NULL;
        return;
    }
    
    if (split[workernum] != NULL) {
        return;
    }
    
    newsplit = (split_t *)calloc(1, sizeof(split_t));
    HPGV_ASSERT(mpiid, newsplit, "Out of memory.", HPGV_ERR_MEM);        
    
    if (workernum == 1) {        
        newsplit->count = 1;
        newsplit->split = (int *)calloc(1, sizeof(int));
        HPGV_ASSERT(mpiid, newsplit->split, "Out of memory.", HPGV_ERR_MEM);
        newsplit->split[0] = 1;
        split[workernum] = newsplit;
        return;
    }        
    
    /* test if it can be divide into 2 */
    count = 2;
    partition[0] = workernum / 2;
    partition[1] = workernum - partition[0];

    lognum[0] = (int)log_2((double) (partition[0]));
    lognum[1] = (int)log_2((double) (partition[1]));
        
    /* if not then divide into 3 */
    if (lognum[0] != lognum[1]) {
        count = 3;
        partition[0] = partition[1] = workernum / 3;
        partition[2] = workernum - partition[0] - partition[1];
        
        /* verify 3 partition */
        for (i = 0; i < 3; i++) {
            lognum[i] = (int)log_2((double) (partition[i]));
        }
            
        if (lognum[0] != lognum[1] || lognum[0] != lognum[2]) {
            HPGV_ABORT(mpiid, "Can not partition a number", HPGV_ERROR);    
        }
    }    
    
    newsplit->count = count;
    newsplit->split = (int *)calloc(count, sizeof(int));
    HPGV_ASSERT(mpiid, newsplit->split, "Out of memory.", HPGV_ERR_MEM);
    
    for (i = 0; i < count; i++) {
        newsplit->split[i] = partition[i];
    }
    
    split[workernum] = newsplit;
    
    /* recursively split */
    for (i = 0; i < count; i++) {
        swapaux_splitworker_ttswap(mpiid, partition[i], split);
    }
}


/**
 * swap_update_wokersplit
 *
 */
static void
swap_update_wokersplit(swap_control_t *swapctrl)
{   
    int mpiid = swapctrl->mpiid;
    int workernum = swapctrl->workernum;
    
    swapctrl->workersplit = (split_t **)realloc(swapctrl->workersplit, 
                                        sizeof(split_t *) * (workernum + 1));
    memset(swapctrl->workersplit, 0, sizeof(split_t *) * (workernum + 1));
    HPGV_ASSERT(mpiid, swapctrl->workersplit, "Out of memory.", HPGV_ERR_MEM);
    
    swapaux_splitworker_ttswap(mpiid, workernum, swapctrl->workersplit);
}


/**
 * swap_update_tree
 *
 */
static void
swap_update_tree(swap_control_t *swapctrl)
{
    int mpiid = swapctrl->mpiid;
    
    if (swapctrl->swaptree != NULL) {
        swaptree_delete_tree(mpiid, swapctrl->swaptree);
        swapctrl->swaptree = NULL;
    }
    
    swapctrl->swaptree = swaptree_new_tree(mpiid, swapctrl->workernum, 
                                           swapctrl->workersplit); 
}


/**
 * swapaux_depth_compare
 *
 */
static int 
swapaux_depth_compare(const void * a, const void * b)
{
    float aa = ((depth_t*)(a))->d;
    float bb = ((depth_t*)(b))->d;
    
    if (aa == bb) {
        return 0;
    }
    
    if (aa < bb) {
        return -1;
    }
    
    return 1;
}


/**
 * swap_update_order
 *
 */
static void
swap_update_order(swap_control_t *swapctrl)
{    
    int mpiid = swapctrl->mpiid;    
    int workernum = swapctrl->workernum;
    int totalprocnum = swapctrl->totalprocnum;
    int root = swapctrl->root;    
    MPI_Comm mpicomm = swapctrl->mpicomm;
    depth_t *dbuf = NULL;
    int i, count, workid;
           
    if (mpiid == root) {
        
        /* gather the depth values */
        dbuf = (depth_t *)calloc(totalprocnum, sizeof(depth_t));
        HPGV_ASSERT(mpiid, dbuf, "Out of memory.", HPGV_ERR_MEM);
                
        MPI_Gather(&(swapctrl->depth), sizeof(depth_t), MPI_BYTE,
                    dbuf, sizeof(depth_t), MPI_BYTE, root, mpicomm);
        
        /* depth sort */
        qsort(dbuf, totalprocnum, sizeof(depth_t), swapaux_depth_compare);        
        
        /* visibility ordering */                      
        swapctrl->workid_2_mpiid = (int *)realloc(swapctrl->workid_2_mpiid,
                                    workernum *sizeof(int));
        HPGV_ASSERT(mpiid, swapctrl->workid_2_mpiid, 
                    "Out of memory.", HPGV_ERR_MEM);
                
        count = 0;
        for (i = 0; i < totalprocnum; i++) {
            if (dbuf[i].mpiid == root && !swapctrl->roothasimage) {
                continue;
            }
            swapctrl->workid_2_mpiid[count++] = dbuf[i].mpiid;
        }
        
        /* broadcast the visibility ordering */
        MPI_Bcast(swapctrl->workid_2_mpiid, sizeof(int) * workernum, MPI_BYTE,
                  root, mpicomm);
        
        
        if (swapctrl->roothasimage == HPGV_TRUE) {
            workid = -1;
            for (i = 0; i < workernum; i++) {
                if (swapctrl->workid_2_mpiid[i] == mpiid) {
                    workid = i;
                    swapctrl->workid = i;
                    break;
                }
            }        
            HPGV_ASSERT(mpiid, workid != -1, 
                        "Can not find myself in the mapping table.", 
                        HPGV_ERROR);
        } else {
            swapctrl->workid = -1;
        }
        
        free(dbuf);
        
    } else {
        
        MPI_Gather(&(swapctrl->depth), sizeof(depth_t), MPI_BYTE,
                    NULL, sizeof(depth_t), MPI_BYTE, root, mpicomm);
        
        swapctrl->workid_2_mpiid = (int *)realloc(swapctrl->workid_2_mpiid,
                                    workernum *sizeof(int));
        HPGV_ASSERT(mpiid, swapctrl->workid_2_mpiid, 
                    "Out of memory.", HPGV_ERR_MEM);
        
        MPI_Bcast(swapctrl->workid_2_mpiid, sizeof(int) * workernum, MPI_BYTE,
                  root, mpicomm);
        
        workid = -1;
        for (i = 0; i < workernum; i++) {            
            if (swapctrl->workid_2_mpiid[i] == mpiid) {
                workid = i;
                swapctrl->workid = i;
                break;
            }
        }
        
        HPGV_ASSERT(mpiid, workid != -1, 
                    "Can not find myself in the mapping table.", HPGV_ERROR);

    }
}

/**
 * swap_update_worker
 *
 */
static void        
swap_update_worker(swap_control_t *swapctrl) 
{
    int mpiid = swapctrl->mpiid;    
    int *proc = NULL;
    int totalprocnum =swapctrl->totalprocnum;
    int i, count;
        
    
    proc = (int *)calloc(totalprocnum, sizeof(int));
    HPGV_ASSERT(mpiid, proc, "Out of memory.", HPGV_ERR_MEM);
    
    /* gather all mpi ids */
    MPI_Allgather(&(mpiid), sizeof(int), MPI_BYTE,
                  proc, sizeof(int), MPI_BYTE, swapctrl->mpicomm);
    
    swapctrl->workers = (int *)realloc(swapctrl->workers, 
                         swapctrl->workernum * sizeof(int));
    HPGV_ASSERT(mpiid, swapctrl->workers, "Out of memory.", HPGV_ERR_MEM);
    
    count = 0;
    for (i = 0; i < totalprocnum; i++) {
        /* skip root if root is not a worker */
        if (proc[i] == swapctrl->root && !swapctrl->roothasimage) {
            continue;
        }
        swapctrl->workers[count++] = proc[i];
    }
    
    free(proc);
}


/**
 * swap_delete_schedule: delete a schedule
 *
 */
static void 
swap_delete_schedule(swap_schedule_t **schedule, uint32_t stagenum)
{
    uint32_t i;
    swap_schedule_t *onesch;
    swap_message_t *onemsg;
    
    for (i = 0; i < stagenum; i++) {
        /* for each stage */
        onesch = schedule[i];
        
        if (onesch == NULL) {
            continue;
        }
        
        /* free the send messages */
        while (onesch->first_send != NULL) {
            onemsg = onesch->first_send;
            onesch->first_send = onemsg->next;
            free(onemsg);            
        }
        
        /* free the recv message */
        while (onesch->first_recv != NULL) {
            onemsg = onesch->first_recv;
            onesch->first_recv = onemsg->next;
            if (onemsg != NULL && onemsg->partial_image !=  NULL) {
                free(onemsg->partial_image);
            }
            free(onemsg);
        }
        
        if (onesch->isendreqs != NULL) {
            free(onesch->isendreqs);
        }           
        
        if (onesch->isendstats != NULL) {
            free(onesch->isendstats);
        }
        
        if (onesch->irecvreqs != NULL) {
            free(onesch->irecvreqs);
        }           
        
        if (onesch->irecvstats != NULL) {
            free(onesch->irecvstats);
        }
        
        free(onesch);
    }    
    
    free(schedule);
}


/**
 * swap_update_schedule
 *
 */
static void
swap_update_schedule(swap_control_t *swapctrl)
{
    int mpiid = swapctrl->mpiid;
    int workid = -1;
    swapnode_t *root = NULL;
    int64_t i, start, end;
        
    if (swapctrl->schedule != NULL) {
        swap_delete_schedule(swapctrl->schedule, swapctrl->stagenum);
    }        
    
    /* allocate memory for the schedule of all stages */
    swapctrl->stagenum = swapctrl->swaptree->maxlevel;
    swapctrl->schedule = (swap_schedule_t **)calloc(swapctrl->stagenum, 
                          sizeof(swap_schedule_t*));
    HPGV_ASSERT(mpiid, swapctrl->schedule, "Out of memory.", HPGV_ERR_MEM);
        
    
    /* travel tree to create schedule at each stage*/
    swaptree_set_schedule(swapctrl, 
                          swapctrl->swaptree->root, 
                          swapctrl->schedule);
    
    /* find the final assignment in the last stage */
    root = swapctrl->swaptree->root;
        
    swapctrl->finalassign = (int64_t *)realloc(swapctrl->finalassign,
                             sizeof(int64_t) * root->workernum * 2);
    HPGV_ASSERT(mpiid, swapctrl->finalassign, "Out of memory.", HPGV_ERR_MEM);
    
    for (i = 0; i < root->workernum; i++) {
        workid = root->assignorder[i];
        
        start = BLOCK_LOW(i, root->workernum, swapctrl->usefulpixels);
        end   = BLOCK_HIGH(i, root->workernum, swapctrl->usefulpixels);
        
        swapctrl->finalassign[workid * 2] = start;
        swapctrl->finalassign[workid * 2 + 1] = end;
    }
    
}


/**
 * swap_control_exchange
 *
 */
static void
swap_control_exchange(swap_control_t *swapctrl, 
                      swap_schedule_t *schedule,
                      int stage)
{
    int mpiid = swapctrl->mpiid;        
    MPI_Comm mpicomm = swapctrl->mpicomm;
    int curstage = schedule->curstage;
    uint8_t *partial_image = swapctrl->partial_image[curstage % 2];

    uint32_t isendnum, irecvnum;
    uint64_t bytesize, byteoffset;
    swap_message_t *inmessage, *outmessage;
    uint8_t *ptr = NULL;
            
    /* Post async receive */
    irecvnum = 0;
    inmessage = schedule->first_recv;
    while (inmessage != NULL) {
        if (inmessage->mpiid != mpiid) {
            bytesize = inmessage->recordcount * inmessage->pixelsize;
 
            /* recv */
            MPI_Irecv(inmessage->partial_image, bytesize, MPI_BYTE,
                      inmessage->mpiid, PARTIAL_IMAGE_MSG, mpicomm,
                      &(schedule->irecvreqs[irecvnum]));

            irecvnum++;            
        }        
        inmessage = inmessage->next;
    }
    
    /* isend all the recordcount information to the destination */
    isendnum = 0;
    outmessage = schedule->first_send;    
    while (outmessage != NULL) {
        bytesize = outmessage->recordcount * outmessage->pixelsize;
        byteoffset = outmessage->offset * outmessage->pixelsize;
        ptr = &(partial_image[byteoffset]);        
                
        /* send */
        MPI_Isend(ptr, bytesize, MPI_BYTE, outmessage->mpiid, 
                  PARTIAL_IMAGE_MSG, mpicomm, &(schedule->isendreqs[isendnum]));
        
        
        isendnum++;                            
        outmessage = outmessage->next;        
    }   
    
    if (irecvnum != 0) {
        MPI_Waitall(irecvnum, schedule->irecvreqs, schedule->irecvstats);
    }
    
    
    if (isendnum != 0) {
        MPI_Waitall(isendnum, schedule->isendreqs, schedule->isendstats);
    }
}


/**
 * swapaux_color_composite
 *
 */
static
void swapaux_color_composite(uint8_t *incolor, uint8_t *outcolor, 
                             int format, int type)
{
    float *fin = NULL, *fout = NULL;
    float alpha = 0;
    
    if (format == HPGV_RGBA && type == HPGV_FLOAT) {
        
        fin = (float *)(incolor);
        fout = (float *)(outcolor);
        
        alpha = 1 - *(fout + 3);
        
        /* red */
        *(fout)     += alpha * (*(fin));        
        /* green */
        *(fout + 1) += alpha * (*(fin + 1));
        /* blue */
        *(fout + 2) += alpha * (*(fin + 2));
        /* alpha */
        *(fout + 3) += alpha * (*(fin + 3));
    }
}



/**
 * swap_control_blend
 *
 */
static void
swap_control_blend(swap_control_t *swapctrl, swap_schedule_t *schedule)                   
{
    swap_message_t *inmessage;
    uint64_t offset, recordcount, i, id;
    int mpiid = swapctrl->mpiid;
    uint32_t pixelsize = swapctrl->pixelsize;
    int curstage = schedule->curstage;

    if (swapctrl->imageformat == HPGV_RGBA && swapctrl->imagetype == HPGV_FLOAT) 
    {
        
        float *outimage = (float *)(swapctrl->partial_image[1 - curstage % 2]);
        
        float *myimage = (float *)(swapctrl->partial_image[curstage % 2]);
        
        float *inimage = NULL;        

        float *compositecolor = NULL;

        float *partialcolor = NULL;

        float alphafactor = 0;
        
        inmessage = schedule->first_recv;

        
    
        while (inmessage != NULL) {
            
            offset = inmessage->offset;
            recordcount = inmessage->recordcount;
                        
            if (inmessage->mpiid != mpiid) {                                
                inimage = (float *)(inmessage->partial_image);
            } else {
                inimage = &(myimage[offset * 4]);
            }

            compositecolor = &(outimage[offset * 4]);
            partialcolor = inimage;

            id = 0;
            for (i = 0; i < recordcount; i++) {
                
                alphafactor = 1.0 - compositecolor[id+3];
                compositecolor[id]   += (alphafactor)*(partialcolor[id]);
                compositecolor[id+1] += (alphafactor)*(partialcolor[id+1]);
                compositecolor[id+2] += (alphafactor)*(partialcolor[id+2]);
                compositecolor[id+3] += (alphafactor)*(partialcolor[id+3]);
                
                id += 4;

                /*
                COLOR_COMPOSITE(&(inimage[i*4]),
                                &(outimage[(offset + i)*4]));
                */
            }
            
            inmessage = inmessage->next;
        }
    }
}

        
/**
 * swap_control_gather
 *
 */
static void
swap_control_gather(swap_control_t *swapctrl)
{
    uint32_t mpiid = swapctrl->mpiid;
    uint32_t workid, sendid;
    uint32_t workernum = swapctrl->workernum;
    uint64_t bytesize, recordcount;
    int64_t start, end;
    uint8_t *partial_image = NULL;  
    uint8_t *final_image = NULL;  
    MPI_Status status;
    uint16_t pixelsize = swapctrl->pixelsize;    
    int curstage = swapctrl->stagenum - 1;
        
    /* I collect all the partial images from the other processors */
    if (mpiid == swapctrl->root) {        
                
        for (workid = 0; workid < workernum; workid++) {
            
            start = swapctrl->finalassign[workid * 2];
            
            if (start == -1) {
                continue;
            }            
            end = swapctrl->finalassign[workid * 2 + 1];            
            
            recordcount = end - start + 1;                                
                  
            bytesize = recordcount * pixelsize;            
            
            final_image = &(swapctrl->final_image[start * pixelsize]);
            
            if (swapctrl->roothasimage && workid == swapctrl->workid) {
                /* copy mine */                  
                memcpy(final_image,
                       &(swapctrl->partial_image[1-curstage%2][start * pixelsize]),
                       recordcount * pixelsize);       
                
            } else {         
                sendid = swapctrl->workid_2_mpiid[workid];                   
                    
                /* recv from others */ 
                MPI_Recv(final_image, bytesize, MPI_CHAR, sendid,
                         FULL_IMAGE_MSG, swapctrl->mpicomm, &status);
                
            }
        }
        
    } else {
        
        workid = swapctrl->workid;
        
        /* I send my partial image to processor 0 */
        start = swapctrl->finalassign[workid * 2];
        
        if (start == -1) {
            return;
        }
            
        end = swapctrl->finalassign[workid * 2 + 1];
        
        
        recordcount = end - start + 1;                                      
        
        bytesize = recordcount * pixelsize;
        
        partial_image
            = &(swapctrl->partial_image[1-curstage%2][start * pixelsize]);
                            
        MPI_Send(partial_image, bytesize, MPI_CHAR, swapctrl->root, 
                 FULL_IMAGE_MSG, swapctrl->mpicomm);        
    }
        
}


/**
 * swap_control_composite
 *
 */
static void
swap_control_composite(swap_control_t *swapctrl)
{   
    
    uint32_t stage;            
        
    /* if there is no worker, we don't need image compositing */
    if (swapctrl->workernum < 1) {
        return;
    }
    
    /* do image compositing */
    for (stage = 0; stage < swapctrl->stagenum; stage++) {
        
        MPI_Barrier(swapctrl->mpicomm);
                
        if (swapctrl->schedule[stage] == NULL) {
            continue;
        }       
        
        /* exchange */
        swap_control_exchange(swapctrl, swapctrl->schedule[stage], stage);
        
        /* blend */
        swap_control_blend(swapctrl, swapctrl->schedule[stage]);
    }       
        
    swap_control_gather(swapctrl);
}


/**
 * swap_control_update
 *
 */
static int
swap_control_update(swap_control_t *swapctrl, 
                    int width, int height, 
                    int format, int type,
                    void *partialpixels,
                    void *finalpixels,
                    float depth, 
                    int root, 
                    MPI_Comm mpicomm,
                    composite_t composite_type) 
{   
    int mpiid = 0;
    
    int workernum = 0;
    int t_workernum = 0;
    int totalprocnum = 0;
    int roothasimage = HPGV_FALSE; 
    int pixelsize = 0;   
    
    update_t lu, gu;
    memset(&lu, 0, sizeof(update_t));
    memset(&gu, 0, sizeof(update_t));
        
    /* my rank */
    MPI_Comm_rank(mpicomm, &mpiid);    
    
    /* pixel size */
    pixelsize = hpgv_formatsize(format) * hpgv_typesize(type);
    
    /* total compositing processor number, including root */
    workernum = 1;
    totalprocnum = 0;
    t_workernum = 0;
    MPI_Allreduce(&workernum, &totalprocnum, 1,                  
                  MPI_INT, MPI_SUM, mpicomm);
    
    /* check if root has image */
    roothasimage = HPGV_FALSE;
    if (mpiid == root && partialpixels != NULL && width != 0 && height != 0) {
        roothasimage = HPGV_TRUE;
    }
    MPI_Bcast(&roothasimage, sizeof(int), MPI_BYTE, root, mpicomm);
    
    if (roothasimage == HPGV_TRUE) {
        t_workernum = totalprocnum;
    } else {
        t_workernum = totalprocnum - 1;
    }
    
    /* update everything if workernum, roothasimage, mpicomm, workernum
       or mpiid changed */
    if (swapctrl->workernum != t_workernum || 
        swapctrl->totalprocnum != totalprocnum ||
        swapctrl->roothasimage != roothasimage ||
        swapctrl->mpicomm != mpicomm ||        
        swapctrl->mpiid != mpiid ||
        swapctrl->root != root ||
        swapctrl->composite_type != composite_type) 
    {
        swapctrl->workernum = t_workernum;
        swapctrl->totalprocnum = totalprocnum;
        swapctrl->roothasimage = roothasimage;
        swapctrl->mpicomm = mpicomm;        
        swapctrl->mpiid = mpiid;
        swapctrl->root = root;
        swapctrl->composite_type = composite_type;
        
        lu.worker = HPGV_TRUE;
        lu.workersplit = HPGV_TRUE;
        lu.tree = HPGV_TRUE;
        lu.order = HPGV_TRUE;
        lu.schedule = HPGV_TRUE;        
    }

         
    /* check if I need to update compositing tree and schedule */            
    if (swapctrl->image_size_x != width || 
        swapctrl->image_size_y != height||
        swapctrl->imageformat != format ||
        swapctrl->imagetype != type)        
    {        
        swapctrl->image_size_x = width;
        swapctrl->image_size_y = height;        
        swapctrl->totalpixels = swapctrl->image_size_x * swapctrl->image_size_y;
        swapctrl->usefulpixels = swapctrl->totalpixels; 
        swapctrl->imageformat = format;
        swapctrl->imagetype = type;
        swapctrl->pixelsize = pixelsize;                
        swapctrl->totalpixelbytes = swapctrl->totalpixels * swapctrl->pixelsize; 
        
        if (!(mpiid == root && roothasimage == HPGV_FALSE)) {
            
            HPGV_ASSERT(mpiid, partialpixels, 
                        "There are no input pixels.", HPGV_ERR_MEM);            
            
            swapctrl->partial_image[0] 
                = (uint8_t *)calloc(swapctrl->totalpixelbytes, sizeof(uint8_t));
        
            swapctrl->partial_image[1] 
                = (uint8_t *)calloc(swapctrl->totalpixelbytes, sizeof(uint8_t));
            
            HPGV_ASSERT(mpiid, 
                        swapctrl->partial_image[0] && swapctrl->partial_image[1],
                        "There are no input pixels.", HPGV_ERR_MEM);
            
            memcpy(swapctrl->partial_image[0], partialpixels, 
               swapctrl->totalpixelbytes);
            
            memcpy(swapctrl->partial_image[1], partialpixels, 
               swapctrl->totalpixelbytes);  
            

        }
        
        
        if (mpiid == root) {
            
            HPGV_ASSERT(mpiid, finalpixels, 
                        "The output buffer is empty.", HPGV_ERR_MEM);
            
            swapctrl->final_image = (uint8_t *)finalpixels;
        }
        
        lu.schedule = HPGV_TRUE;
    }
    
    /* check if I need to update ordering and schedule */
    if (swapctrl->depth.d != depth) {        
        swapctrl->depth.d = depth;
        swapctrl->depth.mpiid = mpiid;        
        lu.order = HPGV_TRUE;
        lu.schedule = HPGV_TRUE;        
    }
    
    /* do nothing if there is no worker */    
    if (swapctrl->workernum == 0) {
        return HPGV_TRUE;
    }
    
    /* check if any one needs update */
    MPI_Allreduce(&lu, &gu, sizeof(update_t ), MPI_BYTE, MPI_BOR, mpicomm);
    
    /* return if we don't need to upate anything */
    if (!gu.worker && !gu.order && !gu.workersplit && 
        !gu.tree && !gu.schedule)
    {
        return HPGV_TRUE;
    }    
    
    
    /* we need to update worker */
    if (gu.worker == HPGV_TRUE) {
        swap_update_worker(swapctrl);
    }    
        
    /* we need to update ordering */
    if (gu.order == HPGV_TRUE) {
        swap_update_order(swapctrl);
    }
    
    /* we need to update cpu partition */
    if (gu.workersplit == HPGV_TRUE) {        
        swap_update_wokersplit(swapctrl);
    }
    
    /* we need to update tree */
    if (gu.tree == HPGV_TRUE) {
        swap_update_tree(swapctrl);
    }
    
    /* we need to update the schedule */
    if (gu.schedule == HPGV_TRUE) {        
        swap_update_schedule(swapctrl);
    } 
       
    
    return HPGV_TRUE;                            
}


/**
 * hpgv_composite_init
 *
 */
int
hpgv_composite_init(MPI_Comm mpicomm)
{
    int mpiid = 0;
    MPI_Comm_rank(mpicomm, &mpiid);
    
    if( theSwapControl == NULL) {
        theSwapControl = (swap_control_t *)calloc(1, sizeof(swap_control_t));
    }    
    
    if (theSwapControl == NULL) {
        HPGV_ERR_MSG(mpiid, "Can not initialize compositing modelue.");
        return HPGV_FALSE;
    }
    
    
    return HPGV_TRUE;
}


/**
 * hpgv_composite
 *
 */
int
hpgv_composite_finalize(MPI_Comm mpicomm)
{
    int mpiid = 0;
    int i;
    MPI_Comm_rank(mpicomm, &mpiid);
    
    swap_control_t *swapctrl = theSwapControl;    
    
    if (swapctrl == NULL) {
        return HPGV_TRUE;
    }

    /* free memory */
    if (swapctrl->workid_2_mpiid != NULL) {
        free(swapctrl->workid_2_mpiid);
    }
    
    if (swapctrl->workers != NULL) {
        free(swapctrl->workers);
    }
    
    if (swapctrl->swaptree != NULL) {
        swaptree_delete_tree(mpiid, swapctrl->swaptree);
    }
    
    if (swapctrl->schedule != NULL) {
        swap_delete_schedule(swapctrl->schedule, swapctrl->stagenum);
    }
    
    if (swapctrl->partial_image[0] != NULL) {
        free(swapctrl->partial_image[0]);
    }
        
    if (swapctrl->partial_image[1] != NULL) {
        free(swapctrl->partial_image[1]);
    }
    
    
    if (swapctrl->finalassign != NULL) {
        free(swapctrl->finalassign);
    }
    
    if (swapctrl->workersplit != NULL) {
        for (i = 0; i < swapctrl->workernum + 1; i++) {
            if (swapctrl->workersplit[i] != NULL) {
                if (swapctrl->workersplit[i]->split != NULL) {
                    free(swapctrl->workersplit[i]->split);
                }
                free(swapctrl->workersplit[i]);
            }
        } 
        
        free(swapctrl->workersplit);
    }
    
    free(swapctrl);
    theSwapControl = NULL;
    
    return HPGV_TRUE;
}


/**
 * hpgv_composite_valid
 *
 */
int
hpgv_composite_valid()
{
    if (theSwapControl == NULL) {
        return HPGV_FALSE;
    }
    
    return HPGV_TRUE;
}


/**
 * hpgv_composite
 *
 */
int  
hpgv_composite(int width, int height, int format, int type, 
               void *partialpixels, void *finalpixels, float depth,
               int root, MPI_Comm mpicomm)
{
    
    
    char context[MAXLINE];
    int mpiid = 0;
    MPI_Comm_rank(mpicomm, &mpiid);
    
    if (theSwapControl == NULL) {
        hpgv_msg(mpiid, root, 
                 "Compositing modelue has not been intialized. Quit compositing.");
        return HPGV_FALSE;
    }
        
    if (width == 0 || height == 0) {
        return HPGV_TRUE;
    }    
                          
    /* upadte swap control */
    swap_control_update(theSwapControl, width, height, format, type, 
                        partialpixels, finalpixels, depth, root, mpicomm,
                        HPGV_TTSWAP);   
                 
    
    /* compositing */
    swap_control_composite(theSwapControl);    
    
    return HPGV_SUCCESS;
}

